/* 
   ABOUT ANTLR:
   [The BSD License]
	Copyright (c) 2010 Terence Parr
	All rights reserved.
	Redistribution and use in source and binary forms, with or without modification, are permitted provided that the 
	following conditions are met:

	    1. Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
	    2. Redistributions in binary form must reproduce the above copyright notice, this 
	      list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
	    3. Neither the name of the author nor the names of its contributors may be used to endorse or promote products 
	      derived from this software without specific prior written permission. 

	THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY 
	EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES 
	OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT 
	SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, 
	SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, 
	PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; 
	OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, 
	STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF 
	THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE. 
 */

/*
 * ABOUT GalaxyGrammar:
 * Created by Timo Hanisch 2010 
 * timohanisch@gmx.de
 * */


grammar Galaxy;
options {
    backtrack=true;
    memoize=true;
    k=2;
    output = AST;
    ASTLabelType=CommonTree;
}
tokens{
	NEGATION;
}

scope Symbols {
	Set types; 
}

@header {
	package de.timoh.galaxyparser;
	import java.util.Set;
	import java.util.HashSet;
	import de.timoh.galaxyparser.typchecking.*;
}

@members {
	private String actualScope = "";
	private String actualType = "";
	private String actualName = "";
	private String actualLine = "";
	
	boolean isTypeName(String name) {
		for (int i = Symbols_stack.size()-1; i>=0; i--) {
			Symbols_scope scope = (Symbols_scope)Symbols_stack.get(i);
			if ( scope.types.contains(name) ) {
				return true;
			}
		}
		return false;
	}
}

@lexer::header {
  package de.timoh.galaxyparser;
}

translation_unit
scope Symbols;
@init {
  $Symbols::types = new HashSet();
}
	: external_declaration+ EOF!
	;


external_declaration
options {k=1;}
	: ( declaration_specifiers? direct_declarator declaration* '{' )=> function_definition
	| declaration
	| struct_specifier
	;

function_definition
scope Symbols; 
@init {
  $Symbols::types = new HashSet();
}
	:	(declaration_specifiers)? direct_declarator {if(Type.isDefinedFunction(actualName)){
													  System.err.println("line "+actualLine+" function with name \'"+actualName+"\' is allready defined" );}else{
													  new Type(actualName,actualType,Type.MAP_TYPE_FUNCTION);
													  if(Main.debug)System.out.println("Neuer function Typ erstellt: "+actualName +" mit typ: "+actualType);
													  }}
		function_block {Type.clearLocal();}
	;
	
function_block
	:	'{'!
		localDeclaration*
		statement*
		'}'!
	;

declaration
scope {
  boolean isTypedef;
}
@init {
  $declaration::isTypedef = false;
}
	: TYPEDEF declaration_specifiers? {$declaration::isTypedef=true;}
	  init_declarator ';' 
	| declaration_specifiers init_declarator? ';'! {if(Type.isDefinedGlobal(actualName)){
													  System.err.println("line "+actualLine+" global variable with name \'"+actualName+"\' is allready defined" );}else{
													  new Type(actualName,actualType,Type.MAP_TYPE_GLOBAL);
													  if(Main.debug)System.out.println("Neuer globaler Typ erstellt: "+actualName +" mit typ: "+actualType);
													  }}
	| INCLUDE STRING_LITERAL
	;
	
localDeclaration
	:	declaration_specifiers_local init_declarator? ';'! {if(Type.isDefinedLocalFunction(actualName)){
													  System.err.println("line "+actualLine+" local variable with name \'"+actualName+"\' is allready defined" );}else{
													  new Type(actualName,actualType,Type.MAP_TYPE_LOCAL);
													  if(Main.debug)System.out.println("Neuer lokaler Typ erstellt: "+actualName +" mit typ: "+actualType);
													  }}
	;
	
declaration_specifiers_local 
	:   (   CONST^
		|   type_specifier^ {actualType = $type_specifier.text;}
        )+
	;

declaration_specifiers
	:   (   storage_class_specifier^
		|   type_specifier^ {actualType = $type_specifier.text;}
        )+
	;

init_declarator
	: direct_declarator ('='^ initializer)?
	;

storage_class_specifier
	: STATIC
	| VOID
	| CONST
	| NATIVE
	;

type_specifier
	:	BOOLEAN
	|	BYTE
	|	CHAR
	|	FIXED_LIT
	|	GAMELINK
	|	COLOR
	|	INT
	|	POINT
	|	STRING
	|	ABILCMD
	|	ACTOR
	|	ACTORSCOPE
	|	BANK
	|	CAMERAINFO
	|	MARKER
	|	ORDER
	|	PLAYERGROUP
	|	REGION
	|	REVEALER
	|	SOUNDLINK
	|	TEXT
	|	TIMER
	|	TRANSMISSIONSOURCE
	|	TRIGGER
	|	UNITFILTER
	|	UNITGROUP
	|	UNITREF
	|	UNIT
	|	WAVEINFO
	|	WAVETARGET
	|	DOODAD
	|	AIFILTER
	|	SOUND
	|	WAVE
	| type_id
	;

type_id
    :   {isTypeName(input.LT(1).getText())}? IDENTIFIER
    	//{System.out.println($IDENTIFIER.text+" is a type");}
    ;

struct_specifier
options {k=3;}
/*scope Symbols; // structs are scopes
@init {
  $Symbols::types = new HashSet();
}*/
	: 
	{actualScope = "struct";}
	STRUCT IDENTIFIER '{'! struct_declaration+ '}'! + ';'! {if(Type.isDefinedStruct($IDENTIFIER.text)){
													  	actualLine =  $IDENTIFIER.line + ":" + $IDENTIFIER.pos;
													  	System.err.println("line "+actualLine+" struct with name \'"+$IDENTIFIER.text+"\' is allready defined" );
													  }else{
														$Symbols::types.add($IDENTIFIER.text);
														new Type($IDENTIFIER.text, $IDENTIFIER.text, Type.MAP_TYPE_STRUCT); 
														if(Main.debug)System.out.println("Neuer struct Typ erstellt: "+actualName);
													  }
													  Type.clearLocal();}
	;

struct_declaration
	: specifier_qualifier_list direct_declarator ('=' expression)? SEMI {if(Type.isDefinedLocalStruct(actualName)){
													  	System.err.println("line "+actualLine+" structlocal with name \'"+actualName+"\' is allready defined" );
													  }else{
														new Type(actualName, actualType, Type.MAP_TYPE_LOCAL); 
														if(Main.debug)System.out.println("Neuer structlocal Typ erstellt: "+actualName);
													  }}
	;
 
specifier_qualifier_list
	: ( type_qualifier | type_specifier )+ {actualType = $type_specifier.text;}
	;

type_qualifier
	: CONST
	;

direct_declarator
	:   
		declarator_suffix*	
		(
			IDENTIFIER
			{
			actualName = $IDENTIFIER.text;
			actualLine =  $IDENTIFIER.line + ":" + $IDENTIFIER.pos;
			if ($declaration.size()>0&& ($declaration::isTypedef)) {
				$Symbols::types.add($IDENTIFIER.text);
				//System.out.println("define type "+$IDENTIFIER.text);
			}
			}
		|	'('! direct_declarator ')'!
		)	
		declarator_suffix*	    
	;

declarator_suffix
	:   '['! constant_expression ']'!
    |   '['! ']'!
    |   '('! parameter_list ')'!
    |   '('! identifier_list ')'!
    |   '('! ')'!
	;

parameter_list
	: parameter_declaration (','! parameter_declaration)*
	;

parameter_declaration
	: declaration_specifiers^ IDENTIFIER
	;

identifier_list
	: IDENTIFIER (','! IDENTIFIER)*
	;

type_name
	: specifier_qualifier_list
	;
	
initializer
	: assignment_expression
	;

initializer_list
	: initializer (','! initializer)*
	;

//expressions

argument_expression_list
	:   assignment_expression (','! assignment_expression)*
	;

additive_expression
	: (multiplicative_expression) ('+'^ multiplicative_expression | '-'^ multiplicative_expression)*
	;

multiplicative_expression
	: (cast_expression) ('*'^ cast_expression | '/'^ cast_expression | '%'^ cast_expression)*
	;

cast_expression
	: '('! type_name ')'! cast_expression
	| unary_expression
	;

unary_expression
	: postfix_expression
	| unary_operator cast_expression
	;

postfix_expression
	:   primary_expression
        (   '['! expression ']'!
        |   '('! ')'!
        |   '('! argument_expression_list ')'!
        |   '.'! IDENTIFIER
        )*
	;

unary_operator
	: '&'^
	| '*'^
	| '+'^
	| '-'^
	| '~'^
	| '!'^
	;

primary_expression
	: IDENTIFIER
	| constant
	| '('! expression ')'!
	;
 
constant
    :   INTEGER
    |	CHARACTER_LITERAL
	|	STRING_LITERAL
    |   FIXED
    |	TRUE
    |	FALSE
    |	NULL
    ;

/////

expression
	: assignment_expression
	;

constant_expression
	: logical_or_expression
	;

assignment_expression
	: lvalue assignment_operator^ assignment_expression
	| logical_or_expression
	;
	
lvalue
	:	unary_expression
	;

assignment_operator
	: '='^
	| '+='^
	| '-='^
	;

logical_or_expression
	: logical_and_expression ('||'^ logical_and_expression)*
	;

logical_and_expression
	: inclusive_or_expression ('&&'^ inclusive_or_expression)*
	;

inclusive_or_expression
	: exclusive_or_expression ('|'^ exclusive_or_expression)*
	;

exclusive_or_expression
	: and_expression ('^'^ and_expression)*
	;

and_expression
	: equality_expression ('&'^ equality_expression)*
	;
equality_expression
	: relational_expression (('=='^|'!='^) relational_expression)*
	;

relational_expression
	: shift_expression (('<'^|'>'^|'<='^|'>='^) shift_expression)*
	;

shift_expression
	: additive_expression (('<<'^|'>>'^) additive_expression)*
	;

//statements

statement
	: block
	| expression_statement
	| selection_statement
	| iteration_statement
	| jump_statement
	;
	
block
	: '{'! statement* '}'!
	;

statement_list
	: statement+
	;

expression_statement
	: ';'!
	| expression ';'!
	;

selection_statement
	: 'if'! '('! expression ')'! statement 
	(options {k=1; backtrack=false;}:'else'! statement )?	
	;

iteration_statement
	: 'while'! '('! expression ')'! statement 
	;

jump_statement
	: 'continue' ';'!
	| 'break' ';'!
	| 'return' ';'!
	| 'return'! expression ';'!
	;

//Lexer

//primitive types
BOOLEAN : 'bool';
BYTE : 'byte';
CHAR : 'char';
FIXED_LIT : 'fixed';
GAMELINK : 'gamelink';
COLOR : 'color';
INT : 'int';
POINT : 'point';
STRING : 'string';

//further types
ABILCMD : 'abilcmd';
ACTOR : 'actor';
ACTORSCOPE : 'actorscope';
AIFILTER : 'aifilter';
BANK : 'bank';
CAMERAINFO : 'camerainfo';
MARKER : 'marker';
ORDER : 'order';
PLAYERGROUP : 'playergroup';
REGION : 'region';
REVEALER : 'revealer';
SOUNDLINK : 'soundlink';
TEXT : 'text';
TIMER : 'timer';
TRIGGER : 'trigger';
TRANSMISSIONSOURCE : 'transmissionsource';
UNITFILTER : 'unitfilter';
UNITGROUP : 'unitgroup';
UNITREF: 'unitref';
UNIT : 'unit';
WAVEINFO : 'waveinfo';
WAVETARGET : 'wavetarget';
DOODAD : 'doodad';
SOUND : 'sound';
WAVE : 'wave';

//Other Keywords
BREAK : 'break';
CONST : 'const';
ELSE : 'else';
FALSE : 'false';
IF : 'if';
INCLUDE : 'include';
NATIVE : 'native';
NULL : 'null';
RETURN : 'return';
STATIC : 'static';
STRUCT : 'struct';
TRUE : 'true';
TYPEDEF : 'typedef';
VOID : 'void';
WHILE : 'while';
SEMI : ';';

IDENTIFIER
	:	LETTER (LETTER|'0'..'9')*
	;
	
fragment
LETTER
	:	'A'..'Z'
	|	'a'..'z'
	|	'_'
	;

CHARACTER_LITERAL
    :   '\'' ( EscapeSequence | ~('\''|'\\') ) '\''
    ;

STRING_LITERAL
    :  '"' ( EscapeSequence | ~('\\'|'"') )* '"'
    ;

INTEGER 
	: DIGIT+
	| HEX_LITERAL
	;
	
fragment
HEX_LITERAL : '0' ('x'|'X') HexDigit+;

fragment
HexDigit : ('0'..'9'|'a'..'f'|'A'..'F') ;

FIXED 
	:	DIGIT+ '.' DIGIT* 
	|	'.' DIGIT+
	;
	
fragment
DIGIT : '0'..'9';

fragment
Exponent : ('e'|'E') ('+'|'-')? ('0'..'9')+ ;

fragment
FloatTypeSuffix : ('f'|'F'|'d'|'D') ;

fragment
EscapeSequence
    :   '\\' ('b'|'t'|'n'|'f'|'r'|'\"'|'\''|'\\')
    |   OctalEscape
    ;

fragment
OctalEscape
    :   '\\' ('0'..'3') ('0'..'7') ('0'..'7')
    |   '\\' ('0'..'7') ('0'..'7')
    |   '\\' ('0'..'7')
    ;

fragment
UnicodeEscape
    :   '\\' 'u' HexDigit HexDigit HexDigit HexDigit
    ;

WS  :  (' '|'\r'|'\t'|'\u000C'|'\n') {$channel=HIDDEN;}
    ;

COMMENT
    :   '/*' ( options {greedy=false;} : . )* '*/' {$channel=HIDDEN;}
    ;

LINE_COMMENT
    : '//' ~('\n'|'\r')* '\r'? '\n' {$channel=HIDDEN;}
    ;
